#include "ModulesInterface.h"
#include "Info.h"
#include <cmath>

using namespace SANC;
using namespace SANCmodulesVar;

void ModulesInterface::init() {
    double gf_gf;

    flags_.iborn = info.Iborn;
    flags_.iqed = info.Iqed;
    flags_.iew = info.Iew;
    switch (info.Scheme)
    {
    case 0:
        flags_.gfscheme = 0;
        break;
    case 1:
        flags_.gfscheme = 2;
        break;
    case 2:
        flags_.gfscheme = 4;
        break;
    }
    flags_.ilin = info.Ilin;
    flags_.ifgg = info.Ifgg;
    flags_.irun = info.Irun;
    flags_.alr = info.Alr;

    polar_.lamep = info.LamPlus;
    polar_.lamem = info.LamMinus;

    masses_.ma = 0.0;
    masses_.mv = info.MZ;
    masses_.mh = info.MH;
    masses_.mz = info.MZ;
    masses_.mw = info.MW;
    masses_.men = info.Men;
    masses_.mel = info.Mel;
    masses_.mmn = info.Mmn;
    masses_.mmo = info.Mmo;
    masses_.mtn = info.Mtn;
    masses_.mta = info.Mta;
    masses_.mup = info.Mup;
    masses_.mdn = info.Mdn;
    masses_.mch = info.Mch;
    masses_.mst = info.Mst;
    masses_.mtp = info.Mtp;
    masses_.mbt = info.Mbt;

    r2masses_.ram2 = masses_.ma*masses_.ma;
    r2masses_.rvm2 = masses_.mv*masses_.mv;
    r2masses_.rhm2 = masses_.mh*masses_.mh;
    r2masses_.rzm2 = masses_.mz*masses_.mz;
    r2masses_.rwm2 = masses_.mw*masses_.mw;
    r2masses_.renm2 = masses_.men*masses_.men;
    r2masses_.relm2 = masses_.mel*masses_.mel;
    r2masses_.rmnm2 = masses_.mmn*masses_.mmn;
    r2masses_.rmom2 = masses_.mmo*masses_.mmo;
    r2masses_.rtnm2 = masses_.mtn*masses_.mtn;
    r2masses_.rtam2 = masses_.mta*masses_.mta;
    r2masses_.rupm2 = masses_.mup*masses_.mup;
    r2masses_.rdnm2 = masses_.mdn*masses_.mdn;
    r2masses_.rchm2 = masses_.mch*masses_.mch;
    r2masses_.rstm2 = masses_.mst*masses_.mst;
    r2masses_.rtpm2 = masses_.mtp*masses_.mtp;
    r2masses_.rbtm2 = masses_.mbt*masses_.mbt;

    c2masses_.cm2 = {r2masses_.ram2, -1.0e-30};
    c2masses_.vm2 = {r2masses_.rvm2, -1.0e-30};
    c2masses_.hm2 = {r2masses_.rhm2, -1.0e-30};
    c2masses_.zm2 = {r2masses_.rzm2, -1.0e-30};
    c2masses_.wm2 = {r2masses_.rwm2, -1.0e-30};
    c2masses_.enm2 = {r2masses_.renm2, -1.0e-30};
    c2masses_.elm2 = {r2masses_.relm2, -1.0e-30};
    c2masses_.upm2 = {r2masses_.rupm2, -1.0e-30};
    c2masses_.dnm2 = {r2masses_.rdnm2, -1.0e-30};
    c2masses_.mnm2 = {r2masses_.rmnm2, -1.0e-30};
    c2masses_.mom2 = {r2masses_.rmom2, -1.0e-30};
    c2masses_.chm2 = {r2masses_.rchm2, -1.0e-30};
    c2masses_.stm2 = {r2masses_.rstm2, -1.0e-30};
    c2masses_.tnm2 = {r2masses_.rtnm2, -1.0e-30};
    c2masses_.tam2 = {r2masses_.rtam2, -1.0e-30};
    c2masses_.tpm2 = {r2masses_.rtpm2, -1.0e-30};
    c2masses_.btm2 = {r2masses_.rbtm2, -1.0e-30};

    cstw_.ctw = masses_.mw/masses_.mz;
    cstw_.stw = sqrt(1.0-cstw_.ctw*cstw_.ctw);
    cstw_.ctw2 = r2masses_.rwm2/r2masses_.rzm2;
    cstw_.stw2 = 1-cstw_.ctw2;
    cstw_.ctw4 = cstw_.ctw2*cstw_.ctw2;
    cstw_.stw4 = cstw_.stw2*cstw_.stw2;
    cstw_.ctw6 = cstw_.ctw4*cstw_.ctw2;
    cstw_.stw6 = cstw_.stw4*cstw_.stw2;

    rmhwz_.rwz = cstw_.ctw2;
    rmhwz_.rhz = r2masses_.rhm2/r2masses_.rzm2;
    rmhwz_.rhw = r2masses_.rhm2/r2masses_.rwm2;
    rmhwz_.renh = r2masses_.renm2/r2masses_.rhm2;
    rmhwz_.relh = r2masses_.relm2/r2masses_.rhm2;
    rmhwz_.ruph = r2masses_.rupm2/r2masses_.rhm2;
    rmhwz_.rdnh = r2masses_.rdnm2/r2masses_.rhm2;
    rmhwz_.rmnh = r2masses_.rmnm2/r2masses_.rhm2;
    rmhwz_.rmoh = r2masses_.rmom2/r2masses_.rhm2;
    rmhwz_.rchh = r2masses_.rchm2/r2masses_.rhm2;
    rmhwz_.rsth = r2masses_.rstm2/r2masses_.rhm2;
    rmhwz_.rtnh = r2masses_.rtnm2/r2masses_.rhm2;
    rmhwz_.rtah = r2masses_.rtam2/r2masses_.rhm2;
    rmhwz_.rtph = r2masses_.rtpm2/r2masses_.rhm2;
    rmhwz_.rbth = r2masses_.rbtm2/r2masses_.rhm2;
    rmhwz_.renz = r2masses_.renm2/r2masses_.rzm2;
    rmhwz_.relz = r2masses_.relm2/r2masses_.rzm2;
    rmhwz_.rupz = r2masses_.rupm2/r2masses_.rzm2;
    rmhwz_.rdnz = r2masses_.rdnm2/r2masses_.rzm2;
    rmhwz_.rmnz = r2masses_.rmnm2/r2masses_.rzm2;
    rmhwz_.rmoz = r2masses_.rmom2/r2masses_.rzm2;
    rmhwz_.rchz = r2masses_.rchm2/r2masses_.rzm2;
    rmhwz_.rstz = r2masses_.rstm2/r2masses_.rzm2;
    rmhwz_.rtnz = r2masses_.rtnm2/r2masses_.rzm2;
    rmhwz_.rtaz = r2masses_.rtam2/r2masses_.rzm2;
    rmhwz_.rtpz = r2masses_.rtpm2/r2masses_.rzm2;
    rmhwz_.rbtz = r2masses_.rbtm2/r2masses_.rzm2;
    rmhwz_.renw = r2masses_.renm2/r2masses_.rwm2;
    rmhwz_.relw = r2masses_.relm2/r2masses_.rwm2;
    rmhwz_.rupw = r2masses_.rupm2/r2masses_.rwm2;
    rmhwz_.rdnw = r2masses_.rdnm2/r2masses_.rwm2;
    rmhwz_.rmnw = r2masses_.rmnm2/r2masses_.rwm2;
    rmhwz_.rmow = r2masses_.rmom2/r2masses_.rwm2;
    rmhwz_.rchw = r2masses_.rchm2/r2masses_.rwm2;
    rmhwz_.rstw = r2masses_.rstm2/r2masses_.rwm2;
    rmhwz_.rtnw = r2masses_.rtnm2/r2masses_.rwm2;
    rmhwz_.rtaw = r2masses_.rtam2/r2masses_.rwm2;
    rmhwz_.rtpw = r2masses_.rtpm2/r2masses_.rwm2;
    rmhwz_.rbtw = r2masses_.rbtm2/r2masses_.rwm2;

    width_.wz = info.Wz;
    width_.ww = info.Ww;
    width_.wh = info.Wh;
    width_.wtp = info.Wtp;
    width_.gz = width_.wz*masses_.mz;
    width_.gw = width_.ww*masses_.mw;
    width_.gh = width_.wh*masses_.mh;
    width_.gtp = width_.wtp*masses_.mtp;

    widthm_.mz2 = {r2masses_.rzm2, -width_.gz};
    widthm_.mz2c = {r2masses_.rzm2, width_.gz};
    widthm_.mw2 = {r2masses_.rwm2, -width_.gw};
    widthm_.mw2c = {r2masses_.rwm2, width_.gw};
    widthm_.mh2 = {r2masses_.rhm2, -width_.gh};
    widthm_.mh2c = {r2masses_.rhm2, width_.gh};
    widthm_.mtp2 = {r2masses_.rtpm2, -width_.gtp};
    widthm_.mtp2c = {r2masses_.rtpm2, width_.gtp};

    scales_.thmu2 = masses_.mw*masses_.mw;
    scales_.tlmu2 = 1.0;
    scales_.omega = info.Ome*info.Ecm/2.0;

    qfi3f_.qf[0] =  0.0;
    qfi3f_.qf[1] = -1.0;
    qfi3f_.qf[2] =  0.0;
    qfi3f_.qf[3] = -1.0;
    qfi3f_.qf[4] =  0.0;
    qfi3f_.qf[5] = -1.0;
    qfi3f_.qf[6] =  2.0/3.0;
    qfi3f_.qf[7] = -1.0/3.0;
    qfi3f_.qf[8] =  2.0/3.0;
    qfi3f_.qf[9] = -1.0/3.0;
    qfi3f_.qf[10] =  2.0/3.0;
    qfi3f_.qf[11] = -1.0/3.0;
    qfi3f_.i3f[0] =  0.5;
    qfi3f_.i3f[1] = -0.5;
    qfi3f_.i3f[2] =  0.5;
    qfi3f_.i3f[3] = -0.5;
    qfi3f_.i3f[4] =  0.5;
    qfi3f_.i3f[5] = -0.5;
    qfi3f_.i3f[6] =  0.5;
    qfi3f_.i3f[7] = -0.5;
    qfi3f_.i3f[8] =  0.5;
    qfi3f_.i3f[9] = -0.5;
    qfi3f_.i3f[10] =  0.5;
    qfi3f_.i3f[11] = -0.5;
    qfi3f_.rmf1[0] = info.Men;
    qfi3f_.rmf1[1] = info.Mel;
    qfi3f_.rmf1[2] = info.Mmn;
    qfi3f_.rmf1[3] = info.Mmo;
    qfi3f_.rmf1[4] = info.Mtn;
    qfi3f_.rmf1[5] = info.Mta;
    qfi3f_.rmf1[6] = info.Mup;
    qfi3f_.rmf1[7] = info.Mdn;
    qfi3f_.rmf1[8] = info.Mch;
    qfi3f_.rmf1[9] = info.Mst;
    qfi3f_.rmf1[10] = info.Mtp;
    qfi3f_.rmf1[11] = info.Mbt;

    qavpm_.qen = qfi3f_.qf[0];
    qavpm_.qel = qfi3f_.qf[1];
    qavpm_.qup = qfi3f_.qf[6];
    qavpm_.qdn = qfi3f_.qf[7];
    qavpm_.qmn = qfi3f_.qf[2];
    qavpm_.qmo = qfi3f_.qf[3];
    qavpm_.qch = qfi3f_.qf[8];
    qavpm_.qst = qfi3f_.qf[9];
    qavpm_.qtn = qfi3f_.qf[4];
    qavpm_.qta = qfi3f_.qf[5];
    qavpm_.qtp = qfi3f_.qf[10];
    qavpm_.qbt = qfi3f_.qf[11];
    qavpm_.aen = qfi3f_.i3f[0];
    qavpm_.ael = qfi3f_.i3f[1];
    qavpm_.aup = qfi3f_.i3f[6];
    qavpm_.adn = qfi3f_.i3f[7];
    qavpm_.amn = qfi3f_.i3f[2];
    qavpm_.amo = qfi3f_.i3f[3];
    qavpm_.ach = qfi3f_.i3f[8];
    qavpm_.ast = qfi3f_.i3f[9];
    qavpm_.atn = qfi3f_.i3f[4];
    qavpm_.ata = qfi3f_.i3f[5];
    qavpm_.atp = qfi3f_.i3f[10];
    qavpm_.abt = qfi3f_.i3f[11];
    qavpm_.ven = qavpm_.aen-2.0*qavpm_.qen*cstw_.stw2;
    qavpm_.vel = qavpm_.ael-2.0*qavpm_.qel*cstw_.stw2;
    qavpm_.vup = qavpm_.aup-2.0*qavpm_.qup*cstw_.stw2;
    qavpm_.vdn = qavpm_.adn-2.0*qavpm_.qdn*cstw_.stw2;
    qavpm_.vmn = qavpm_.amn-2.0*qavpm_.qmn*cstw_.stw2;
    qavpm_.vmo = qavpm_.amo-2.0*qavpm_.qmo*cstw_.stw2;
    qavpm_.vch = qavpm_.ach-2.0*qavpm_.qch*cstw_.stw2;
    qavpm_.vst = qavpm_.ast-2.0*qavpm_.qst*cstw_.stw2;
    qavpm_.vtn = qavpm_.atn-2.0*qavpm_.qtn*cstw_.stw2;
    qavpm_.vta = qavpm_.ata-2.0*qavpm_.qta*cstw_.stw2;
    qavpm_.vtp = qavpm_.atp-2.0*qavpm_.qtp*cstw_.stw2;
    qavpm_.vbt = qavpm_.abt-2.0*qavpm_.qbt*cstw_.stw2;
    qavpm_.vpaen = qavpm_.ven+qavpm_.aen;
    qavpm_.vmaen = qavpm_.ven-qavpm_.aen;
    qavpm_.vpael = qavpm_.vel+qavpm_.ael;
    qavpm_.vmael = qavpm_.vel-qavpm_.ael;
    qavpm_.vpaup = qavpm_.vup+qavpm_.aup;
    qavpm_.vmaup = qavpm_.vup-qavpm_.aup;
    qavpm_.vpadn = qavpm_.vdn+qavpm_.adn;
    qavpm_.vmadn = qavpm_.vdn-qavpm_.adn;
    qavpm_.vpamn = qavpm_.vmn+qavpm_.amn;
    qavpm_.vmamn = qavpm_.vmn-qavpm_.amn;
    qavpm_.vpamo = qavpm_.vmo+qavpm_.amo;
    qavpm_.vmamo = qavpm_.vmo-qavpm_.amo;
    qavpm_.vpach = qavpm_.vch+qavpm_.ach;
    qavpm_.vmach = qavpm_.vch-qavpm_.ach;
    qavpm_.vpast = qavpm_.vst+qavpm_.ast;
    qavpm_.vmast = qavpm_.vst-qavpm_.ast;
    qavpm_.vpatn = qavpm_.vtn+qavpm_.atn;
    qavpm_.vmatn = qavpm_.vtn-qavpm_.atn;
    qavpm_.vpata = qavpm_.vta+qavpm_.ata;
    qavpm_.vmata = qavpm_.vta-qavpm_.ata;
    qavpm_.vpatp = qavpm_.vtp+qavpm_.atp;
    qavpm_.vmatp = qavpm_.vtp-qavpm_.atp;
    qavpm_.vpabt = qavpm_.vbt+qavpm_.abt;
    qavpm_.vmabt = qavpm_.vbt-qavpm_.abt;
    qavpm_.dfen = (qavpm_.ven-qavpm_.aen)/qavpm_.aen;
    qavpm_.dfel = (qavpm_.vel-qavpm_.ael)/qavpm_.ael;
    qavpm_.dfup = (qavpm_.vup-qavpm_.aup)/qavpm_.aup;
    qavpm_.dfdn = (qavpm_.vdn-qavpm_.adn)/qavpm_.adn;
    qavpm_.dfmn = (qavpm_.vmn-qavpm_.amn)/qavpm_.amn;
    qavpm_.dfmo = (qavpm_.vmo-qavpm_.amo)/qavpm_.amo;
    qavpm_.dfch = (qavpm_.vch-qavpm_.ach)/qavpm_.ach;
    qavpm_.dfst = (qavpm_.vst-qavpm_.ast)/qavpm_.ast;
    qavpm_.dftn = (qavpm_.vtn-qavpm_.atn)/qavpm_.atn;
    qavpm_.dfta = (qavpm_.vta-qavpm_.ata)/qavpm_.ata;
    qavpm_.dftp = (qavpm_.vtp-qavpm_.atp)/qavpm_.atp;
    qavpm_.dfbt = (qavpm_.vbt-qavpm_.abt)/qavpm_.abt;

    consts_.pi = 3.141592653589793;
    consts_.sr2 = sqrt(2.0);
    consts_.alphas = 0.1176;
    consts_.alphai = info.Alphai;
    consts_.gf = 1.16637e-5;
    gf_gf = consts_.gf;
    consts_.gf = consts_.pi/consts_.alphai/sqrt(2.0)/cstw_.stw2/cstw_.ctw2/r2masses_.rzm2;
    nc_.cfscheme = consts_.gf/gf_gf;
    nc_.cfprime = nc_.cfscheme;
    if (flags_.gfscheme == 0) {
    nc_.cfscheme = 1.0;
    nc_.cfprime = 1.0;
    } else if (flags_.gfscheme == 2) {
    consts_.gf = gf_gf;
    consts_.alphai = consts_.pi/consts_.gf/consts_.sr2/cstw_.stw2/cstw_.ctw2/r2masses_.rzm2;
    nc_.cfscheme = 1.0;
    } else if (flags_.gfscheme == 4) {
    consts_.alphai = info.AlphaMZi;
    consts_.gf = consts_.pi/consts_.alphai/sqrt(2.0)/cstw_.stw2/cstw_.ctw2/r2masses_.rzm2;
    nc_.cfprime = consts_.alphai/info.Alphai;
    nc_.cfscheme = 1.0/nc_.cfprime;
    }
    consts_.alpha = 1.0/consts_.alphai;
    consts_.e = sqrt(4.0*consts_.pi*consts_.alpha);
    consts_.g = sqrt(consts_.gf/consts_.sr2*8.0*r2masses_.rzm2*cstw_.ctw2);
    consts_.gs = sqrt(4.0*consts_.pi*consts_.alphas);
    consts_.conhc = 0.38937966e9;
    consti_.i_ = {0.0, 1.0};
    nc_.nc = 1.0;
    nc_.fc = 3.0;
    nc_.cfl = 1.0;
    nc_.cfq = 3.0;
}

ModulesInterface::ModulesInterface(Info & _info):info(_info)
{
    init();
}

void SANCmodulesVar::setMomenta(SANC::Event & event) {
    event.P(1).toArr(momenta_.pg1);
    event.P(2).toArr(momenta_.pg2);
    event.P(3).toArr(momenta_.pg3);
    event.P(4).toArr(momenta_.pg4);
    event.P(5).toArr(momenta_.pg5);
}