#include "Process.h"
#include "General.h"
#include "Info.h"
#include "ModulesInterface.h"
using namespace SANC;
using namespace SANCmodules;

Process::Process(Info &_info) : info(_info) {
  idup.resize(5);
  idup[0] = 0;
}

void Process::init() {
  Mf1_2 = pow2(Mf1);
  Mf2_2 = pow2(Mf2);
  Mf3_2 = pow2(Mf3);
  Mf4_2 = pow2(Mf4);
  sqS = info.Ecm;
  S = pow2(sqS);
  E1 = (S + Mf1_2 - Mf2_2) / (2. * sqS);
  E2 = (S - Mf1_2 + Mf2_2) / (2. * sqS);
  beta1 = sqrt(1 - Mf1_2 / pow2(E1));
  beta2 = sqrt(1 - Mf2_2 / pow2(E2));

  dm = (Mf1 - Mf2) * (Mf1 + Mf2) * (Mf4 - Mf3) * (Mf4 + Mf3);
  sm = Mf1_2 + Mf2_2 + Mf3_2 + Mf4_2;
  event.particles.clear();
  for (auto const &id : idup) {
    event.particles.push_back(Particle(id));
  }
  event.particles.push_back(Particle(22));

  // std::cout<<"-----------------------------------------------------"<<'\n';
  std::cout << " Process: " << procName << '\n';
  std::cout << " Citation: " << arXiv << '\n';
  std::cout << " ====================================================" << '\n';
}

Process &SANC::ProcessFactory(Info &info) {
  switch (info.Pid) {
  case 101:
    return *(new Process_12_12__12_12(info));
    break;
  case 102:
    return *(new Process_12_12__02_04(info));
    break;
  case 103:
    return *(new Process_12_12__16_16(info));
    break;
  case 104:
    return *(new Process_12_12__20_20(info));
    break;
  default:
    std::cerr << "Error: Unknown process " << info.Pid << std::endl;
    exit(1);
  }
}

Process_12_12__12_12::Process_12_12__12_12(Info &_info) : Process(_info) {
  procName = "e+e- -> e-e+";
  arXiv = "Phys.Rev.D 98, 013001; arXiv:1801.00125";
  Mf1 = info.Mel;
  Mf2 = info.Mel;
  Mf3 = info.Mel;
  Mf4 = info.Mel;
  idup[1] = -11;
  idup[2] = 11;
  idup[3] = 11;
  idup[4] = -11;
  kDimBorn = 2;
  kDimVirt = 2;
  kDimHard = 5;
  init();
}

void Process_12_12__12_12::born(double *Qs, double *Qt, double *Qu, double *Born, double *Soft,
                                double *Hard) {
  nc_br_12_12__12_12_(Qs, Qt, Qu, Born, Soft, Hard);
}

void Process_12_12__12_12::virt(double *S, double *T, double *U, double *Sigma) {
  nc_si_12_12__12_12_(S, T, U, Sigma);
}

void Process_12_12__12_12::hard(double *S, double *Spr, double *Phi3, double *Eta3, double *Eta5,
                                double *Hard) {
  nc_ha_12_12__12_12_(S, Spr, Phi3, Eta3, Eta5, Hard);
}

Process_12_12__16_16::Process_12_12__16_16(Info &_info) : Process(_info) {
  procName = "e+e- -> mu-mu+";
  arXiv = "";
  Mf1 = info.Mel;
  Mf2 = info.Mel;
  Mf3 = info.Mmo;
  Mf4 = info.Mmo;
  idup[1] = -11;
  idup[2] = 11;
  idup[3] = 13;
  idup[4] = -13;
  kDimBorn = 2;
  kDimVirt = 2;
  kDimHard = 5;
  init();
}

void Process_12_12__16_16::born(double *Qs, double *Qt, double *Qu, double *Born, double *Soft,
                                double *Hard) {
  nc_br_12_12__16_16_(Qs, Qt, Qu, Born, Soft, Hard);
}

void Process_12_12__16_16::virt(double *S, double *T, double *U, double *Sigma) {
  nc_si_12_12__16_16_(S, T, U, Sigma);
}

void Process_12_12__16_16::hard(double *S, double *Spr, double *Phi3, double *Eta3, double *Eta5,
                                double *Hard) {
  nc_ha_12_12__16_16_(S, Spr, Phi3, Eta3, Eta5, Hard);
}

Process_12_12__20_20::Process_12_12__20_20(Info &_info) : Process(_info) {
  procName = "e+e- -> tau-tau+";
  arXiv = "";
  Mf1 = info.Mel;
  Mf2 = info.Mel;
  Mf3 = info.Mta;
  Mf4 = info.Mta;
  idup[1] = -11;
  idup[2] = 11;
  idup[3] = 15;
  idup[4] = -15;
  kDimBorn = 2;
  kDimVirt = 2;
  kDimHard = 5;
  init();
}

void Process_12_12__20_20::born(double *Qs, double *Qt, double *Qu, double *Born, double *Soft,
                                double *Hard) {
  nc_br_12_12__20_20_(Qs, Qt, Qu, Born, Soft, Hard);
}

void Process_12_12__20_20::virt(double *S, double *T, double *U, double *Sigma) {
  nc_si_12_12__20_20_(S, T, U, Sigma);
}

void Process_12_12__20_20::hard(double *S, double *Spr, double *Phi3, double *Eta3, double *Eta5,
                                double *Hard) {
  nc_ha_12_12__20_20_(S, Spr, Phi3, Eta3, Eta5, Hard);
}

Process_12_12__02_04::Process_12_12__02_04(Info &_info) : Process(_info) {
  procName = "e+e- -> ZH";
  arXiv = "Phys.Rev.D 100, 073002; arXiv:1812.10965";
  Mf1 = info.Mel;
  Mf2 = info.Mel;
  Mf3 = info.MZ;
  Mf4 = info.MH;
  idup[1] = -11;
  idup[2] = 11;
  idup[3] = 23;
  idup[4] = 25;
  kDimBorn = 2;
  kDimVirt = 2;
  kDimHard = 5;
  init();
}

void Process_12_12__02_04::born(double *Qs, double *Qt, double *Qu, double *Born, double *Soft,
                                double *Hard) {
  nc_br_12_12__02_04_(Qs, Qt, Qu, Born, Soft, Hard);
}

void Process_12_12__02_04::virt(double *S, double *T, double *U, double *Sigma) {
  nc_si_12_12__02_04_(S, T, U, Sigma);
}

void Process_12_12__02_04::hard(double *S, double *Spr, double *Phi3, double *Eta3, double *Eta5,
                                double *Hard) {
  nc_ha_12_12__02_04_(S, Spr, Phi3, Eta3, Eta5, Hard);
}
