#include "ReneSANCe_ee.h"
#include "TRandom3.h"
#include <iostream>
#include <iomanip>
#include <sys/stat.h>
#include <complex>
using namespace std;
typedef complex<double> double_complex;
#include "clooptools.h"
#include "TFoam.h"
#include "Integrand.h"
#include "TFile.h"
#include "TTree.h"
#include "Info.h"
#include "Process.h"
#include "ModulesInterface.h"
#include "Time.h"
#include "DataSaver.h"


extern void LHEInit(void);

using namespace SANC;

#define SP12 setprecision(12) << setw(21)
#define SW18 setw(18)
#define SW3  setw(3)
#define SW6  setw(6)


ReneSANCe_ee::ReneSANCe_ee(Settings & _cfg, Info & _info, Process & _proc, ModulesInterface & _mod, Time & _time, DataSaver & _ds):
cfg(_cfg),timing(_time),info(_info),proc(_proc),modules(_mod),ds(_ds)
{
    PseRan=new TRandom3();
    PseRan->SetSeed(info.RanSeed);
    ffini();
}

void ReneSANCe_ee::generateGrid()
{

}

void ReneSANCe_ee::generateEvents()
{
    TFoam *FoamBorn,*FoamVirt,*FoamHard;
    TFoamIntegrand *Density0,*Density1,*Density2;
    TFile *fileBorn,*fileVirt,*fileHard;
    char bornName[30], bornNameLhe[30], bornNameRoot[30], bornNameDat[30];
    char loopName[30], loopNameLhe[30], loopNameRoot[30], loopNameDat[30];
    bool bBorn,bVirt,bHard;
    bBorn = true;
    bVirt = !info.Iborn && !info.ForceOffSV && (info.Iew || info.Iqed);
    bHard = !info.Iborn && !info.ForceOffH && info.Iqed;
    sprintf(bornName, "SANCeeLO");
    sprintf(loopName, "SANCeeNLO");
    if (info.Alr == 1 || info.Alr == 3) {
	sprintf(bornName, "%sdif", bornName);
	sprintf(loopName, "%sdif", loopName);
    } else if (info.Alr == 2 || info.Alr == 4) {
	sprintf(bornName, "%ssum", bornName);
	sprintf(loopName, "%ssum", loopName);
    }
    sprintf(bornNameLhe, "Results/%s.lhe",bornName);
    sprintf(bornNameRoot, "Results/%s.root",bornName);
    sprintf(bornNameDat, "Results/%s.dat",bornName);
    sprintf(loopNameLhe, "Results/%s.lhe",loopName);
    sprintf(loopNameRoot, "Results/%s.root",loopName);
    sprintf(loopNameDat, "Results/%s.dat",loopName);

    timing.startClock();
    if (info.IEXPLBORN) {
	FoamBorn = new TFoam("FoamBorn");   // Create Simulator
	Density0 = new Integrand(info, proc, 0);   // Create integrand function
	FoamBorn->SetkDim(proc.kDimBorn);
	FoamBorn->SetnCells(info.nCellsBorn);
	FoamBorn->SetnSampl(info.nSamplBorn);
	FoamBorn->SetnBin(info.nBinBorn);
	FoamBorn->SetEvPerBin(info.EvPerBinBorn);
	FoamBorn->SetOptRej(info.OptRej);
	FoamBorn->SetOptDrive(info.OptDrive);
	FoamBorn->SetMaxWtRej(info.MaxWtRejBorn);
	FoamBorn->SetChat(info.Chat);
	FoamBorn->SetRho(Density0);
	FoamBorn->SetPseRan(PseRan);
	FoamBorn->SetInhiDiv(1, 1);
	FoamBorn->Initialize();
	TFile RootBornFile("FoamBorn.root","RECREATE","Born");
	FoamBorn->Write("FoamBorn");    // Writing Foam on the disk
    } else {
	fileBorn = new TFile("FoamBorn.root"); // connect disk file
	FoamBorn = (TFoam*)fileBorn->Get("FoamBorn"); // find object
	Density0 = new Integrand(info, proc, 0);   // Create integrand function
	FoamBorn->SetPseRan(PseRan);
	FoamBorn->SetRho(Density0);
	FoamBorn->SetMaxWtRej(info.MaxWtRejBorn);
	FoamBorn->LinkCells(); // restore pointers of the binary tree of cells
	FoamBorn->GetTotDim();
    }
    timing.stopClock();
    timing.ReportClock("Born initialization");
	FoamBorn->GetPrimary(MCresult);
	if (MCresult<1.e-300) {
        bBorn = false;
        bVirt = false;
    }
    if (bVirt) {
	timing.startClock();
	if (info.IEXPLVIRT) {
	    FoamVirt = new TFoam("FoamVirt");   // Create Simulator
	    Density1 = new Integrand(info, proc, 1);   // Create integrand function
	    FoamVirt->SetkDim(proc.kDimVirt);
	    FoamVirt->SetnCells(info.nCellsVirt);
	    FoamVirt->SetnSampl(info.nSamplVirt);
	    FoamVirt->SetnBin(info.nBinVirt);
	    FoamVirt->SetEvPerBin(info.EvPerBinVirt);
	    FoamVirt->SetOptRej(info.OptRej);
	    FoamVirt->SetOptDrive(info.OptDrive);
	    FoamVirt->SetMaxWtRej(info.MaxWtRejVirt);
	    FoamVirt->SetChat(info.Chat);
	    FoamVirt->SetRho(Density1);
	    FoamVirt->SetPseRan(PseRan);
	    FoamVirt->SetInhiDiv(1, 1);
	    FoamVirt->Initialize();
	    TFile RootVirtFile("FoamVirt.root","RECREATE","Virt");
	    FoamVirt->Write("FoamVirt");    // Writing Foam on the disk
	} else {
	    fileVirt = new TFile("FoamVirt.root"); // connect disk file
	    FoamVirt = (TFoam*)fileVirt->Get("FoamVirt"); // find object
	    Density1 = new Integrand(info, proc, 1);   // Create integrand function
	    FoamVirt->SetPseRan(PseRan);
	    FoamVirt->SetRho(Density1);
	    FoamVirt->SetMaxWtRej(info.MaxWtRejVirt);
	    FoamVirt->LinkCells(); // restore pointers of the binary tree of cells
	    FoamVirt->GetTotDim();
	}
	timing.stopClock();
	timing.ReportClock("virt initialization");
    }
    if (bHard) {
	timing.startClock();
	if (info.IEXPLHARD) {
	    FoamHard = new TFoam("FoamHard");   // Create Simulator
	    Density2 = new Integrand(info, proc, 2);   // Create integrand function
	    FoamHard->SetkDim(proc.kDimHard);
	    FoamHard->SetnCells(info.nCellsHard);
	    FoamHard->SetnSampl(info.nSamplHard);
	    FoamHard->SetnBin(info.nBinHard);
	    FoamHard->SetEvPerBin(info.EvPerBinHard);
	    FoamHard->SetOptRej(info.OptRej);
	    FoamHard->SetOptDrive(info.OptDrive);
	    FoamHard->SetMaxWtRej(info.MaxWtRejHard);
	    FoamHard->SetChat(info.Chat);
	    FoamHard->SetRho(Density2);
	    FoamHard->SetPseRan(PseRan);
	    FoamHard->SetInhiDiv(4, 1);
	    FoamHard->Initialize();
	    TFile RootHardFile("FoamHard.root","RECREATE","Hard");
	    FoamHard->Write("FoamHard");    // Writing Foam on the disk
	} else {
	    fileHard = new TFile("FoamHard.root"); // connect disk file
	    FoamHard = (TFoam*)fileHard->Get("FoamHard"); // find object
	    Density2 = new Integrand(info, proc, 2);   // Create integrand function
	    FoamHard->SetPseRan(PseRan);
	    FoamHard->SetRho(Density2);
	    FoamHard->SetMaxWtRej(info.MaxWtRejHard);
	    FoamHard->LinkCells(); // restore pointers of the binary tree of cells
	    FoamHard->GetTotDim();
	}
	timing.stopClock();
	timing.ReportClock("hard initialization");
    FoamHard->GetPrimary(MCresult);
	if (MCresult<1.e-300) { bHard = false; }
    }
    timing.startClock();
    cout << "====== Initialization done, entering MC loop" << '\n';
//--------------------------------------------------------------------------------------------------
    double *MCvectBorn = new double[proc.kDimBorn];   // vector generated in the MC run
    double *MCvectVirt = new double[proc.kDimVirt];   // vector generated in the MC run
    double *MCvectHard = new double[proc.kDimHard];   // vector generated in the MC run
//--------------------------------------------------------------------------------------------------
    int iBornPWt1  = 0, iBornNWt1  = 0;
    int iBornPWtN1 = 0, iBornNWtN1 = 0;
    int iVirtPWt1  = 0, iVirtNWt1  = 0;
    int iVirtPWtN1 = 0, iVirtNWtN1 = 0;
    int iHardPWt1  = 0, iHardNWt1  = 0;
    int iHardPWtN1 = 0, iHardNWtN1 = 0;
//--------------------------------------------------------------------------------------------------
    mkdir("Results", S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
//--------------------------------------------------------------------------------------------------
	if (bBorn) {
		output_brn.open(bornNameLhe);
		ds.setOutput(&output_brn);
		ds.start();
	//--------------------------------------------------------------------------------------------------
		TFile *fborn = new TFile(bornNameRoot,"recreate");
		TTree *tborn = new TTree("events_born","born events");
		Double_t weight_born[1], px_born[3], py_born[3], pz_born[3], E_born[3];
		tborn->Branch("weight_born", weight_born, "weight_born[1]/D");
		tborn->Branch("px_born", px_born, "px_born[3]/D");
		tborn->Branch("py_born", py_born, "py_born[3]/D");
		tborn->Branch("pz_born", pz_born, "pz_born[3]/D");
		tborn->Branch("E_born", E_born, "E_born[3]/D");
	//--------------------------------------------------------------------------------------------------
		for (loop = 0; loop < info.NevBorn; loop++) {
		FoamBorn->MakeEvent(); // generate Born MC event
		FoamBorn->GetMCvect(MCvectBorn);
		MCwt = FoamBorn->GetMCwt();
		xwgtup = proc.event.weight();
		if (MCwt == 1.0) {
			weight_born[0] = xwgtup;
			px_born[0] = proc.event.P(3).px();
			px_born[1] = proc.event.P(4).px();
			px_born[2] = proc.event.P(5).px();
			py_born[0] = proc.event.P(3).py();
			py_born[1] = proc.event.P(4).py();
			py_born[2] = proc.event.P(5).py();
			pz_born[0] = proc.event.P(3).pz();
			pz_born[1] = proc.event.P(4).pz();
			pz_born[2] = proc.event.P(5).pz();
			E_born[0]  = proc.event.P(3).e();
			E_born[1]  = proc.event.P(4).e();
			E_born[2]  = proc.event.P(5).e();
			if (info.PrintLHE == 1) ds.writeEvent();
			if (info.PrintROOT == 1) tborn->Fill();
			if (xwgtup > 0.0) iBornPWt1++;
			if (xwgtup < 0.0) iBornNWt1++;
		} else {
			if (xwgtup > 0.0) iBornPWtN1++;
			if (xwgtup < 0.0) iBornNWtN1++;
		}
		}
		fborn->Write();
		timing.stopClock();
		timing.ReportClock("Born generation");
    	FoamBorn->GetIntegMC(MCresult, MCerror); // get MC integral
	}
    branch[0] = MCresult;
    branch_err[0] = MCerror;
//--------------------------------------------------------------------------------------------------
    output_tot.open(loopNameLhe);
	ds.setOutput(&output_tot);
	ds.start();
//--------------------------------------------------------------------------------------------------
    if (bVirt) {
	timing.startClock();
	for (loop = 0; loop < info.NevVirt; loop++) {
	    FoamVirt->MakeEvent(); // generate Virt MC event
	    FoamVirt->GetMCvect(MCvectVirt);
	    MCwt = FoamVirt->GetMCwt();
	}
	timing.stopClock();
	timing.ReportClock("virt generation");
	FoamVirt->GetIntegMC(MCresult, MCerror);  // get MC integral
	branch[1] = MCresult;
	branch_err[1] = MCerror;
    }
//--------------------------------------------------------------------------------------------------
    if (bHard) {
	timing.startClock();
	for (loop = 0; loop < info.NevHard; loop++) {
	    FoamHard->MakeEvent();   // generate Hard MC event
	    FoamHard->GetMCvect(MCvectHard);
	    MCwt = FoamHard->GetMCwt();
	}
	timing.stopClock();
	timing.ReportClock("hard generation");
	FoamHard->GetIntegMC(MCresult, MCerror);  // get MC integral
	branch[2] = MCresult;
	branch_err[2] = MCerror;
    }
//--------------------------------------------------------------------------------------------------
    info.SigmaBorn = branch[0];
    info.SigmaVirt = branch[1];
    info.SigmaHard = branch[2];
    info.SigmaTot  = info.SigmaVirt+info.SigmaHard;
//--------------------------------------------------------------------------------------------------
    cout << "-------------------" << '\n';
    cout << "Branching weights:" << '\n';
    cout << "Born = " << branch[0] << " +- " << branch_err[0] << '\n';
    cout << "Virt = " << branch[1] << " +- " << branch_err[1] << '\n';
    cout << "Hard = " << branch[2] << " +- " << branch_err[2] << '\n';
    cout << "-------------------" << '\n';
//--------------------------------------------------------------------------------------------------
	if (info.SigmaTot>1.e-300) {
        double rBr;
        double BrVirt = info.SigmaVirt/info.SigmaTot;
//--------------------------------------------------------------------------------------------------
        TFile *ftot = new TFile(loopNameRoot,"recreate");
        TTree *ttot = new TTree("events_tot","tot events");
        Double_t weight_tot[1], px_tot[3], py_tot[3], pz_tot[3], E_tot[3];
        ttot->Branch("weight_tot", weight_tot, "weight_tot[1]/D");
        ttot->Branch("px_tot", px_tot, "px_tot[3]/D");
        ttot->Branch("py_tot", py_tot, "py_tot[3]/D");
        ttot->Branch("pz_tot", pz_tot, "pz_tot[3]/D");
        ttot->Branch("E_tot", E_tot, "E_tot[3]/D");
//--------------------------------------------------------------------------------------------------
        TFile *fhardvar = new TFile("Results/hardvar.root","recreate");
        TTree *thardvar = new TTree("events_hardvar","tot events");
        Double_t weight_hardvar[1], hardvar[5];
        thardvar->Branch("weight_hardvar", weight_hardvar, "weight_hardvar[1]/D");
        thardvar->Branch("hardvar", hardvar, "hardvar[5]/D");
//--------------------------------------------------------------------------------------------------
        timing.startClock();
        for (loop = 0; loop < info.NevTot; loop++) {
        rBr = (double) random()/RAND_MAX;
//--------------------------------------------------------------------------------------------------
        if (rBr <= BrVirt) {
            FoamVirt->MakeEvent();   // generate Virt MC event
            FoamVirt->GetMCvect(MCvectVirt);
            MCwt = FoamVirt->GetMCwt();
            xwgtup = proc.event.weight();
            if (MCwt == 1.0) {
            weight_tot[0] = xwgtup;
            px_tot[0] = proc.event.P(3).px();
            px_tot[1] = proc.event.P(4).px();
            px_tot[2] = proc.event.P(5).px();
            py_tot[0] = proc.event.P(3).py();
            py_tot[1] = proc.event.P(4).py();
            py_tot[2] = proc.event.P(5).py();
            pz_tot[0] = proc.event.P(3).pz();
            pz_tot[1] = proc.event.P(4).pz();
            pz_tot[2] = proc.event.P(5).pz();
            E_tot[0]  = proc.event.P(3).e();
            E_tot[1]  = proc.event.P(4).e();
            E_tot[2]  = proc.event.P(5).e();
            if (info.PrintROOT == 1) ttot->Fill();
            if (info.PrintLHE == 1) ds.writeEvent();
            if (xwgtup > 0.0) iVirtPWt1++;
            if (xwgtup < 0.0) iVirtNWt1++;
            } else {
            if (xwgtup > 0.0) iVirtPWtN1++;
            if (xwgtup < 0.0) iVirtNWtN1++;
            }
        }
//--------------------------------------------------------------------------------------------------
        else {
            FoamHard->MakeEvent();   // generate Hard MC event
            FoamHard->GetMCvect(MCvectHard);
            MCwt = FoamHard->GetMCwt();
            xwgtup = proc.event.weight();
            if (MCwt == 1.0) {
            weight_tot[0] = xwgtup;
            px_tot[0] = proc.event.P(3).px();
            px_tot[1] = proc.event.P(4).px();
            px_tot[2] = proc.event.P(5).px();
            py_tot[0] = proc.event.P(3).py();
            py_tot[1] = proc.event.P(4).py();
            py_tot[2] = proc.event.P(5).py();
            pz_tot[0] = proc.event.P(3).pz();
            pz_tot[1] = proc.event.P(4).pz();
            pz_tot[2] = proc.event.P(5).pz();
            E_tot[0]  = proc.event.P(3).e();
            E_tot[1]  = proc.event.P(4).e();
            E_tot[2]  = proc.event.P(5).e();
            weight_hardvar[0] = xwgtup;
            for (int i = 0; i < 5; i++) hardvar[i] = MCvectHard[i];
            if (info.PrintLHE == 1) ds.writeEvent();
            if (info.PrintROOT == 1) ttot->Fill();
            if (info.PrintHard == 1) thardvar->Fill();
            if (xwgtup > 0.0) iHardPWt1++;
            if (xwgtup < 0.0) iHardNWt1++;
            } else {
            if (xwgtup > 0.0) iHardPWtN1++;
            if (xwgtup < 0.0) { iHardNWtN1++; /*MCwt = -MCwt;*/ }
            }
        }
        }
        ftot->Write();
        fhardvar->Write();
        timing.stopClock();
        timing.ReportClock("Total generation");
//--------------------------------------------------------------------------------------------------
        output_brn << "</LesHouchesEvents>" << endl;
        output_tot << "</LesHouchesEvents>" << endl;
        output_brn.close();
        output_tot.close();
    }
//--------------------------------------------------------------------------------------------------
    double BornF = 1.0;
    double VirtF = 1.0;
    double HardF = 1.0;
    if ((iBornPWt1+iBornNWt1+iBornPWtN1+iBornNWtN1) > 0) BornF = (double) (iBornPWt1-iBornNWt1+iBornPWtN1-iBornNWtN1)/(iBornPWt1+iBornNWt1+iBornPWtN1+iBornNWtN1);
    if ((iVirtPWt1+iVirtNWt1+iVirtPWtN1+iVirtNWtN1) > 0) VirtF = (double) (iVirtPWt1-iVirtNWt1+iVirtPWtN1-iVirtNWtN1)/(iVirtPWt1+iVirtNWt1+iVirtPWtN1+iVirtNWtN1);
    if ((iHardPWt1+iHardNWt1+iHardPWtN1+iHardNWtN1) > 0) HardF = (double) (iHardPWt1-iHardNWt1+iHardPWtN1-iHardNWtN1)/(iHardPWt1+iHardNWt1+iHardPWtN1+iHardNWtN1);
//--------------------------------------------------------------------------------------------------
    cout << "====== Events generated, entering Finalize" << '\n';
    Double_t IntNorm, Errel;
    FoamBorn->Finalize(IntNorm, Errel);
    double SigmaBornAbs(0.),SigmaVirtAbs(0.),SigmaHardAbs(0.);
    double ErrBorn(0.),ErrVirt(0.),ErrHard(0.);
    double aveWt,wtMax,sigma;
//--------------------------------------------------------------------------------------------------
    FoamBorn->GetIntegMC(MCresult, MCerror);  // get MC integral
    info.SigmaBorn = BornF*MCresult;
    ErrBorn = MCerror;
    SigmaBornAbs = MCresult;
//--------------------------------------------------------------------------------------------------
    if (bVirt) {
        FoamVirt->Finalize(IntNorm, Errel);
        FoamVirt->GetIntegMC(MCresult, MCerror);  // get MC integral
        info.SigmaVirt = VirtF*MCresult;
        ErrVirt = MCerror;
        SigmaVirtAbs = MCresult;
    }
//--------------------------------------------------------------------------------------------------
    if (bHard) {
        FoamHard->Finalize(IntNorm, Errel);
        FoamHard->GetIntegMC(MCresult, MCerror);  // get MC integral
        FoamHard->GetWtParams(0.0001,aveWt,wtMax,sigma);
        info.SigmaHard = HardF*MCresult;
        ErrHard = MCerror;
        SigmaHardAbs = MCresult;
    }
//--------------------------------------------------------------------------------------------------
    info.SigmaTot = info.SigmaVirt+info.SigmaHard;
    double SigmaDel = info.SigmaVirt+info.SigmaHard-info.SigmaBorn;
    double SigmaTotAbs = SigmaVirtAbs+SigmaHardAbs;
    double ErrTot = sqrt(ErrVirt*ErrVirt+ErrHard*ErrHard);
    double ErrDel = sqrt(ErrVirt*ErrVirt+ErrHard*ErrHard+ErrBorn*ErrBorn);
    double Delta = SigmaDel/info.SigmaBorn;
    double ErrDelta = Delta*sqrt(ErrDel/SigmaDel*ErrDel/SigmaDel+ErrBorn/info.SigmaBorn*ErrBorn/info.SigmaBorn);
    ofstream sigma_brn;
    ofstream sigma_tot;
    sigma_brn.open(bornNameDat);
    sigma_tot.open(loopNameDat);
    sigma_brn << setprecision(10) << SigmaBornAbs << '\n';
    sigma_tot << setprecision(10) << SigmaTotAbs << '\n';
    sigma_brn.close();
    sigma_tot.close();
//--------------------------------------------------------------------------------------------------
    delete [] MCvectBorn;
    delete [] MCvectVirt;
    delete [] MCvectHard;
//--------------------------------------------------------------------------------------------------
    cout << "================================================================" << '\n';
    cout << "        MCresult (Born)  = " << info.SigmaBorn << " +- " << ErrBorn << " RelErr = " << ErrBorn/info.SigmaBorn << '\n';
	cout << "        MCresult (B+S+V) = " << info.SigmaVirt << " +- " << ErrVirt << " RelErr = " << ErrVirt/info.SigmaVirt << '\n';
	cout << "        MCresult (Hard)  = " << info.SigmaHard << " +- " << ErrHard << " RelErr = " << ErrHard/info.SigmaHard << '\n';
    cout << "        MCresult (Tot)   = " << info.SigmaTot << " +- " << ErrTot << " RelErr = " << ErrTot/info.SigmaTot << '\n';
    cout << "        MCresult (Delta) = " << Delta << " +- " << ErrDelta << " RelErr = " << ErrDelta/Delta << '\n';
    cout << "================================================================" << '\n';
    cout << "WtMax(for 1e-4 accuracy goal) = "<<wtMax<< '\n';
	cout << "================================================================" << '\n';
    cout << "Total number of Born events:  " << iBornPWt1 + iBornPWtN1 << " pos. " << iBornNWt1 + iBornNWtN1 << " neg." << '\n';
    cout << "Total number of B+S+V events: " << iVirtPWt1 + iVirtPWtN1 << " pos. " << iVirtNWt1 + iVirtNWtN1 << " neg." << '\n';
    cout << "Total number of Hard events:  " << iHardPWt1 + iHardPWtN1 << " pos. " << iHardNWt1 + iHardNWtN1 << " neg." << '\n';
    cout << "================================================================" << '\n';
    cout << "Number of Born events with Wt > 1:  " << iBornPWtN1 << " pos. " << iBornNWtN1 << " neg." << '\n';
    cout << "Number of B+S+V events with Wt > 1: " << iVirtPWtN1 << " pos. " << iVirtNWtN1 << " neg." << '\n';
    cout << "Number of Hard events with Wt > 1:  " << iHardPWtN1 << " pos. " << iHardNWtN1 << " neg." << '\n';
    cout << "================================================================" << '\n';
    cout << "Relative number of Born events with Wt > 1:  " << (double) (iBornPWtN1+iBornNWtN1)/(iBornPWt1+iBornNWt1+iBornPWtN1+iBornNWtN1) << '\n';
    cout << "Relative number of B+S+V events with Wt > 1: " << (double) (iVirtPWtN1+iVirtNWtN1)/(iVirtPWt1+iVirtNWt1+iVirtPWtN1+iVirtNWtN1) << '\n';
    cout << "Relative number of Hard events with Wt > 1:  " << (double) (iHardPWtN1+iHardNWtN1)/(iHardPWt1+iHardNWt1+iHardPWtN1+iHardNWtN1) << '\n';
    cout << "***** End of Program *****" << '\n';
}

void ReneSANCe_ee::generateEventsNB()
{
    TFoam *FoamBorn,*FoamLoop;
    TFoamIntegrand *Density0,*Density3;
    TFile *fileBorn,*fileLoop;
    timing.startClock();
    if (info.IEXPLBORN) {
	FoamBorn = new TFoam("FoamBorn");   // Create Simulator
	Density0 = new Integrand(info, proc, 0);   // Create integrand function
	FoamBorn->SetkDim(proc.kDimBorn);
	FoamBorn->SetnCells(info.nCellsBorn);
	FoamBorn->SetnSampl(info.nSamplBorn);
	FoamBorn->SetnBin(info.nBinBorn);
	FoamBorn->SetEvPerBin(info.EvPerBinBorn);
	FoamBorn->SetOptRej(info.OptRej);
	FoamBorn->SetOptDrive(info.OptDrive);
	FoamBorn->SetMaxWtRej(info.MaxWtRejBorn);
	FoamBorn->SetChat(info.Chat);
	FoamBorn->SetRho(Density0);
	FoamBorn->SetPseRan(PseRan);
	FoamBorn->SetInhiDiv(1, 1);
	FoamBorn->Initialize();
	TFile RootBornFile("FoamBorn.root","RECREATE","Born");
	FoamBorn->Write("FoamBorn");    // Writing Foam on the disk
    } else {
	fileBorn = new TFile("FoamBorn.root"); // connect disk file
	FoamBorn = (TFoam*)fileBorn->Get("FoamBorn"); // find object
	Density0 = new Integrand(info, proc, 0);   // Create integrand function
	FoamBorn->SetPseRan(PseRan);
	FoamBorn->SetRho(Density0);
	FoamBorn->SetMaxWtRej(info.MaxWtRejBorn);
	FoamBorn->LinkCells(); // restore pointers of the binary tree of cells
	FoamBorn->GetTotDim();
    }
    timing.stopClock();
    timing.ReportClock("Born initialization");
    if (!info.Iborn && (info.Iew || info.Iqed)) {
	timing.startClock();
	if (info.IEXPLHARD) {
	    FoamLoop = new TFoam("FoamLoop");   // Create Simulator
	    Density3 = new Integrand(info, proc, 3);   // Create integrand function
	    FoamLoop->SetkDim(proc.kDimHard+1);
	    FoamLoop->SetnCells(info.nCellsHard);
	    FoamLoop->SetnSampl(info.nSamplHard);
	    FoamLoop->SetnBin(info.nBinHard);
	    FoamLoop->SetEvPerBin(info.EvPerBinHard);
	    FoamLoop->SetOptRej(info.OptRej);
	    FoamLoop->SetOptDrive(info.OptDrive);
	    FoamLoop->SetMaxWtRej(info.MaxWtRejHard);
	    FoamLoop->SetChat(info.Chat);
	    FoamLoop->SetRho(Density3);
	    FoamLoop->SetPseRan(PseRan);
	    FoamLoop->SetInhiDiv(5,1);
	    double xDiv[1];
	    xDiv[0]=0.5;
	    FoamLoop->SetXdivPRD(5,1,xDiv);
	    FoamLoop->Initialize();
	    TFile RootVirtFile("FoamLoop.root","RECREATE","Loop");
	    FoamLoop->Write("FoamLoop");    // Writing Foam on the disk
	} else {
	    fileLoop = new TFile("FoamLoop.root"); // connect disk file
	    FoamLoop = (TFoam*)fileLoop->Get("FoamLoop"); // find object
	    Density3 = new Integrand(info, proc, 3);   // Create integrand function
	    FoamLoop->SetPseRan(PseRan);
	    FoamLoop->SetRho(Density3);
	    FoamLoop->SetMaxWtRej(info.MaxWtRejHard);
	    FoamLoop->LinkCells(); // restore pointers of the binary tree of cells
	    FoamLoop->GetTotDim();
	}
	timing.stopClock();
	timing.ReportClock("loop initialization");
    }
    timing.startClock();
    cout << "====== Initialization done, entering MC loop" << '\n';
//--------------------------------------------------------------------------------------------------
    double *MCvectBorn = new double[proc.kDimBorn];   // vector generated in the MC run
    double *MCvectLoop = new double[proc.kDimHard+1];   // vector generated in the MC run
//--------------------------------------------------------------------------------------------------
    int iBornPWt1  = 0, iBornNWt1  = 0;
    int iBornPWtN1 = 0, iBornNWtN1 = 0;
    int iLoopPWt1  = 0, iLoopNWt1  = 0;
    int iLoopPWtN1 = 0, iLoopNWtN1 = 0;
//--------------------------------------------------------------------------------------------------
    mkdir("Results", S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
    output_brn.open("Results/SANCeeLO.lhe");
    output_tot.open("Results/SANCeeNLO.lhe");
//--------------------------------------------------------------------------------------------------
	ds.setOutput(&output_brn);
	ds.start();
//--------------------------------------------------------------------------------------------------
    TFile *fborn = new TFile("Results/SANCeeLO.root","recreate");
    TTree *tborn = new TTree("events_born","born events");
    Double_t weight_born[1], px_born[3], py_born[3], pz_born[3], E_born[3];
    tborn->Branch("weight_born", weight_born, "weight_born[1]/D");
    tborn->Branch("px_born", px_born, "px_born[3]/D");
    tborn->Branch("py_born", py_born, "py_born[3]/D");
    tborn->Branch("pz_born", pz_born, "pz_born[3]/D");
    tborn->Branch("E_born", E_born, "E_born[3]/D");
//--------------------------------------------------------------------------------------------------
    for (loop = 0; loop < info.NevBorn; loop++) {
	FoamBorn->MakeEvent();   // generate Born MC event
	FoamBorn->GetMCvect(MCvectBorn);
	MCwt = FoamBorn->GetMCwt();
	xwgtup = proc.event.weight();
	if (MCwt == 1.0) {
	    weight_born[0] = xwgtup;
	    px_born[0] = proc.event.P(3).px();
	    px_born[1] = proc.event.P(4).px();
	    px_born[2] = proc.event.P(5).px();
	    py_born[0] = proc.event.P(3).py();
	    py_born[1] = proc.event.P(4).py();
	    py_born[2] = proc.event.P(5).py();
	    pz_born[0] = proc.event.P(3).pz();
	    pz_born[1] = proc.event.P(4).pz();
	    pz_born[2] = proc.event.P(5).pz();
	    E_born[0]  = proc.event.P(3).e();
	    E_born[1]  = proc.event.P(4).e();
	    E_born[2]  = proc.event.P(5).e();
	    if (info.PrintLHE == 1) ds.writeEvent();
	    if (info.PrintROOT == 1) tborn->Fill();
	    if (xwgtup > 0.0) iBornPWt1++;
	    if (xwgtup < 0.0) iBornNWt1++;
	} else {
	    if (xwgtup > 0.0) iBornPWtN1++;
	    if (xwgtup < 0.0) iBornNWtN1++;
	}
    }
    fborn->Write();
    timing.stopClock();
    timing.ReportClock("Born generation");
//--------------------------------------------------------------------------------------------------
	ds.setOutput(&output_tot);
	ds.start();
//--------------------------------------------------------------------------------------------------
    TFile *ftot = new TFile("Results/SANCeeNLO.root","recreate");
    TTree *ttot = new TTree("events_tot","tot events");
    Double_t weight_tot[1], px_tot[3], py_tot[3], pz_tot[3], E_tot[3];
    ttot->Branch("weight_tot", weight_tot, "weight_tot[1]/D");
    ttot->Branch("px_tot", px_tot, "px_tot[3]/D");
    ttot->Branch("py_tot", py_tot, "py_tot[3]/D");
    ttot->Branch("pz_tot", pz_tot, "pz_tot[3]/D");
    ttot->Branch("E_tot", E_tot, "E_tot[3]/D");
//--------------------------------------------------------------------------------------------------
    timing.startClock();
    for (loop = 0; loop < info.NevTot; loop++) {
	FoamLoop->MakeEvent();   // generate Hard MC event
	FoamLoop->GetMCvect(MCvectLoop);
	MCwt = FoamLoop->GetMCwt();
	xwgtup = proc.event.weight();
	if (MCwt == 1.0) {
	    weight_tot[0] = xwgtup;
	    px_tot[0] = proc.event.P(3).px();
	    px_tot[1] = proc.event.P(4).px();
	    px_tot[2] = proc.event.P(5).px();
	    py_tot[0] = proc.event.P(3).py();
	    py_tot[1] = proc.event.P(4).py();
	    py_tot[2] = proc.event.P(5).py();
	    pz_tot[0] = proc.event.P(3).pz();
	    pz_tot[1] = proc.event.P(4).pz();
	    pz_tot[2] = proc.event.P(5).pz();
	    E_tot[0]  = proc.event.P(3).e();
	    E_tot[1]  = proc.event.P(4).e();
	    E_tot[2]  = proc.event.P(5).e();
	    if (info.PrintLHE == 1) ds.writeEvent();
	    if (info.PrintROOT == 1) ttot->Fill();
	    if (xwgtup > 0.0) iLoopPWt1++;
	    if (xwgtup < 0.0) iLoopNWt1++;
	} else {
	    if (xwgtup > 0.0) iLoopPWtN1++;
	    if (xwgtup < 0.0) iLoopNWtN1++;
        }
    }
    ftot->Write();
    timing.stopClock();
    timing.ReportClock("Total generation");
//--------------------------------------------------------------------------------------------------
    output_brn << "</LesHouchesEvents>" << endl;
    output_tot << "</LesHouchesEvents>" << endl;
    output_brn.close();
    output_tot.close();
//--------------------------------------------------------------------------------------------------
    double BornF = 1.;
    double HardF = 1.;
    if ((iBornPWt1+iBornNWt1+iBornPWtN1+iBornNWtN1) > 0) BornF = (double) (iBornPWt1-iBornNWt1+iBornPWtN1-iBornNWtN1)/(iBornPWt1+iBornNWt1+iBornPWtN1+iBornNWtN1);
    if ((iLoopPWt1+iLoopNWt1+iLoopPWtN1+iLoopNWtN1) > 0) HardF = (double) (iLoopPWt1-iLoopNWt1+iLoopPWtN1-iLoopNWtN1)/(iLoopPWt1+iLoopNWt1+iLoopPWtN1+iLoopNWtN1);
//--------------------------------------------------------------------------------------------------
    cout << "====== Events generated, entering Finalize" << '\n';
    Double_t IntNorm, Errel;
    double aveWt,wtMax,sigma;
    FoamBorn->Finalize(IntNorm, Errel);
    FoamLoop->Finalize(IntNorm, Errel);
    FoamLoop->GetWtParams(0.0001,aveWt,wtMax,sigma);
//--------------------------------------------------------------------------------------------------
    FoamBorn->GetIntegMC(MCresult, MCerror);  // get MC integral
    info.SigmaBorn = BornF*MCresult;
    double ErrBorn = MCerror;
    double SigmaBornAbs = MCresult;
    cout << "================================================================" << '\n';
    cout << "        MCresult (Born) = " << info.SigmaBorn << " +- " << ErrBorn << " RelErr = " << ErrBorn/info.SigmaBorn << '\n';
//--------------------------------------------------------------------------------------------------
    FoamLoop->GetIntegMC(MCresult, MCerror);  // get MC integral
    info.SigmaHard = HardF*MCresult;
    double ErrHard = MCerror;
    double SigmaHardAbs = MCresult;
    cout << "        MCresult (Hard) = " << info.SigmaHard << " +- " << ErrHard << " RelErr = " << ErrHard/info.SigmaHard << '\n';
//--------------------------------------------------------------------------------------------------
    info.SigmaTot = info.SigmaHard;
    double SigmaDel = info.SigmaHard-info.SigmaBorn;
    double SigmaTotAbs = SigmaHardAbs;
    double ErrTot = sqrt(ErrHard*ErrHard);
    double ErrDel = sqrt(ErrHard*ErrHard+ErrBorn*ErrBorn);
    double Delta = SigmaDel/info.SigmaBorn;
    double ErrDelta = Delta*sqrt(ErrDel/SigmaDel*ErrDel/SigmaDel+ErrBorn/info.SigmaBorn*ErrBorn/info.SigmaBorn);
    cout << "        MCresult (Tot)  = " << info.SigmaTot << " +- " << ErrTot << " RelErr = " << ErrTot/info.SigmaTot << '\n';
    cout << "        MCresult (Delta)  = " << Delta << " +- " << ErrDelta << " RelErr = " << ErrDelta/Delta << '\n';
    cout << "================================================================" << '\n';
    cout << "WtMax(esp= 0.0001) = "<<wtMax<< '\n';
    ofstream sigma_brn;
    ofstream sigma_tot;
    sigma_brn.open("Results/SANCeeLO.dat");
    sigma_tot.open("Results/SANCeeNLO.dat");
    sigma_brn << setprecision(10) << SigmaBornAbs << '\n';
    sigma_tot << setprecision(10) << SigmaTotAbs << '\n';
    sigma_brn.close();
    sigma_tot.close();
//--------------------------------------------------------------------------------------------------
    delete [] MCvectBorn;
    delete [] MCvectLoop;
//--------------------------------------------------------------------------------------------------
    cout << "Total number of Born events: " << iBornPWt1 + iBornPWtN1 << " pos. " << iBornNWt1 + iBornNWtN1 << " neg." << '\n';
    cout << "Total number of Loop events: " << iLoopPWt1 + iLoopPWtN1 << " pos. " << iLoopNWt1 + iLoopNWtN1 << " neg." << '\n';
    cout << "================================================================" << '\n';
    cout << "Number of Born events with Wt > 1: " << iBornPWtN1 << " pos. " << iBornNWtN1 << " neg." << '\n';
    cout << "Number of Hard events with Wt > 1: " << iLoopPWtN1 << " pos. " << iLoopNWtN1 << " neg." << '\n';
    cout << "================================================================" << '\n';
    cout << "Relative number of Born events with Wt > 1: " << (double) (iBornPWtN1+iBornNWtN1)/(iBornPWt1+iBornNWt1+iBornPWtN1+iBornNWtN1) << '\n';
    cout << "Relative number of Hard events with Wt > 1: " << (double) (iLoopPWtN1+iLoopNWtN1)/(iLoopPWt1+iLoopNWt1+iLoopPWtN1+iLoopNWtN1) << '\n';
    cout << "***** End of Program *****" << '\n';
}
