//--------------------------------------------------------------------------------------------------
#include<fstream>
#include<cmath>
#include<iomanip>
#include<sstream>
#include <sys/stat.h>
#include <complex>
using namespace std;
#include "Riostream.h"
#include "TFile.h"
#include "TFoam.h"
#include "TCanvas.h"
#include "TH3.h"
#include "TH2.h"
#include "TH1.h"
#include "TMath.h"
#include "TFoamIntegrand.h"
#include "TFDISTR.h"
#include "TRandom3.h"
#include "TTree.h"
#include "TStyle.h"
#include "foam_init_nobr.h"
typedef complex<double> double_complex;
#include "clooptools.h"
#define SP12 setprecision(12) << setw(21)
#define SW18 setw(18)
#define SW3  setw(3)
#define SW6  setw(6)
ofstream output_born;
ofstream output_loop;
extern void LHEInit(void);
extern void ParamInit(void);
//--------------------------------------------------------------------------------------------------
extern struct LHEinit {
    int idbmup[2];
    double ebmup[2];
    int pdfgup[2],pdfsup[2],idwtup,nprup;
    double xsecup[10],xerrup[10],xmaxup[10];
    int lprup[10];
} heprup_;
//--------------------------------------------------------------------------------------------------
extern struct LHEevent {
    int nup, idprup;
    double xwgtup, scalup, aqedup, aqcdup;
    int idup[10], istup[10], mothup[10][2], icolup[10][2];
    double pup[10][5], vtimup[10], spinup[10];
} hepeup_;
//--------------------------------------------------------------------------------------------------
void print_born_init()
{
    LHEInit();
    output_born << "<LesHouchesEvents version=\"1.0\">" << endl;
    output_born << "<!--" << endl;
    output_born << "File generated with SANCee MC generator" << endl;
    output_born << "-->" << endl;
    output_born << "<init>" << endl;
    output_born << SW6 << heprup_.idbmup[0]<< SW6 << heprup_.idbmup[1]
               << SP12 << heprup_.ebmup[0] << SP12 << heprup_.ebmup[1]
               << SW6 << heprup_.pdfgup[0] << SW6 << heprup_.pdfgup[1]
               << SW6 << heprup_.pdfsup[0] << SW6 << heprup_.pdfsup[1]
               << SW6 << heprup_.idwtup << SW6 << heprup_.nprup << endl;
    output_born << SP12 << heprup_.xsecup[0] << SP12 << heprup_.xerrup[0] << SP12 << heprup_.xmaxup[0]
               << SW18 << heprup_.lprup[0] << endl;
    output_born << "</init>" << endl;
}
//--------------------------------------------------------------------------------------------------
void print_loop_init()
{
    LHEInit();
    output_loop << "<LesHouchesEvents version=\"1.0\">" << endl;
    output_loop << "<!--" << endl;
    output_loop << "File generated with SANCee MC generator" << endl;
    output_loop << "-->" << endl;
    output_loop << "<init>" << endl;
    output_loop << SW6 << heprup_.idbmup[0]<< SW6 << heprup_.idbmup[1]
               << SP12 << heprup_.ebmup[0] << SP12 << heprup_.ebmup[1]
               << SW6 << heprup_.pdfgup[0] << SW6 << heprup_.pdfgup[1]
               << SW6 << heprup_.pdfsup[0] << SW6 << heprup_.pdfsup[1]
               << SW6 << heprup_.idwtup << SW6 << heprup_.nprup << endl;
    output_loop << SP12 << heprup_.xsecup[0] << SP12 << heprup_.xerrup[0] << SP12 << heprup_.xmaxup[0]
               << SW18 << heprup_.lprup[0] << endl;
    output_loop << "</init>" << endl;
}
//--------------------------------------------------------------------------------------------------
void print_born_lhe()
{
    output_born << "<event>" << endl
    << SW6 << hepeup_.nup << SW6 << hepeup_.idprup << SW6 << hepeup_.xwgtup << SP12 << hepeup_.scalup
    << SP12 << hepeup_.aqedup << SP12 << hepeup_.aqcdup << endl;
    for (int i = 0; i < hepeup_.nup; i++) {
        output_born << SW6 << hepeup_.idup[i] << SW6 << hepeup_.istup[i] << SW6 << hepeup_.mothup[i][0]
        << SW6 << hepeup_.mothup[i][1] << SW6 << hepeup_.icolup[i][0] << SW6 << hepeup_.icolup[i][1]
        << SP12 << hepeup_.pup[i][0] << SP12 << hepeup_.pup[i][1] << SP12 << hepeup_.pup[i][2]
        << SP12 << hepeup_.pup[i][3] << SP12 << hepeup_.pup[i][4] << SW6 << hepeup_.vtimup[i]
        << SW6 << hepeup_.spinup[i] << endl;
    }
    output_born << "</event>" << endl;
}
//--------------------------------------------------------------------------------------------------
void print_loop_lhe()
{
    output_loop << "<event>" << endl
    << SW6 << hepeup_.nup << SW6 << hepeup_.idprup << SW6 << hepeup_.xwgtup << SP12 << hepeup_.scalup
    << SP12 << hepeup_.aqedup << SP12 << hepeup_.aqcdup << endl;
    for (int i = 0; i < hepeup_.nup; i++) {
        output_loop << SW6 << hepeup_.idup[i] << SW6 << hepeup_.istup[i] << SW6 << hepeup_.mothup[i][0]
        << SW6 << hepeup_.mothup[i][1] << SW6 << hepeup_.icolup[i][0] << SW6 << hepeup_.icolup[i][1]
        << SP12 << hepeup_.pup[i][0] << SP12 << hepeup_.pup[i][1] << SP12 << hepeup_.pup[i][2]
        << SP12 << hepeup_.pup[i][3] << SP12 << hepeup_.pup[i][4] << SW6 << hepeup_.vtimup[i]
        << SW6 << hepeup_.spinup[i] << endl;
    }
    output_loop << "</event>" << endl;
}
//--------------------------------------------------------------------------------------------------
int main(void) {
    ffini();
    ParamInit();
    double xwgtup;
    long   loop;
    double MCresult,MCerror,MCwt;
    TRandom *PseRan   = new TRandom3();
    PseRan->SetSeed(RanSeed);
#if IEXPLBORN == 1
    TFoam *FoamBorn = new TFoam("FoamBorn");   // Create Simulator
    TFoamIntegrand *Density0 = new TFDISTR(0);   // Create integrand function
    FoamBorn->SetkDim(           kDimBorn);
    FoamBorn->SetnCells(       nCellsBorn);
    FoamBorn->SetnSampl(       nSamplBorn);
    FoamBorn->SetnBin(           nBinBorn);
    FoamBorn->SetEvPerBin(   EvPerBinBorn);
    FoamBorn->SetOptRej(           OptRej);
    FoamBorn->SetOptDrive(       OptDrive);
    FoamBorn->SetMaxWtRej(   MaxWtRejBorn);
    FoamBorn->SetChat(               Chat);
    FoamBorn->SetRho(Density0);
    FoamBorn->SetPseRan(PseRan);
    FoamBorn->Initialize();
    TFile RootBornFile("FoamBornNobr.root","RECREATE","Born");
    FoamBorn->Write("FoamBorn");    // Writing Foam on the disk
#endif
#if IEXPLLOOP == 1
    TFoam *FoamLoop = new TFoam("FoamLoop");   // Create Simulator
    TFoamIntegrand *Density3 = new TFDISTR(3);   // Create integrand function
    FoamLoop->SetkDim(           kDimLoop);
    FoamLoop->SetnCells(       nCellsLoop);
    FoamLoop->SetnSampl(       nSamplLoop);
    FoamLoop->SetnBin(           nBinLoop);
    FoamLoop->SetEvPerBin(   EvPerBinLoop);
    FoamLoop->SetOptRej(           OptRej);
    FoamLoop->SetOptDrive(       OptDrive);
    FoamLoop->SetMaxWtRej(   MaxWtRejLoop);
    FoamLoop->SetChat(               Chat);
    FoamLoop->SetRho(Density3);
    FoamLoop->SetPseRan(PseRan);
    FoamLoop->Initialize();
    TFile RootLoopFile("FoamLoopNobr.root","RECREATE","Loop");
    FoamLoop->Write("FoamLoop");    // Writing Foam on the disk
#endif
    cout << "====== Initialization done, entering MC loop" << endl;
#if IEXPLBORN == 0
    TFile fileBorn("FoamBornNobr.root"); // connect disk file
    TFoam *FoamBorn = (TFoam*)fileBorn.Get("FoamBorn"); // find object
    TFoamIntegrand *Density0 = new TFDISTR(0);   // Create integrand function
    FoamBorn->SetPseRan(PseRan);
    FoamBorn->SetRho(Density0);
    FoamBorn->SetMaxWtRej(       MaxWtRejBorn);
    FoamBorn->LinkCells(); // restore pointers of the binary tree of cells
    FoamBorn->GetTotDim();
#endif
#if IEXPLLOOP == 0
    TFile fileLoop("FoamLoopNobr.root"); // connect disk file
    TFoam *FoamLoop = (TFoam*)fileLoop.Get("FoamLoop"); // find object
    TFoamIntegrand *Density3 = new TFDISTR(3);   // Create integrand function
    FoamLoop->SetPseRan(PseRan);
    FoamLoop->SetRho(Density3);
    FoamLoop->SetMaxWtRej(       MaxWtRejLoop);
    FoamLoop->LinkCells(); // restore pointers of the binary tree of cells
    FoamLoop->GetTotDim();
#endif
//--------------------------------------------------------------------------------------------------
    double *MCvectBorn = new double[kDimBorn];   // vector generated in the MC run
    double *MCvectLoop = new double[kDimLoop];   // vector generated in the MC run
//--------------------------------------------------------------------------------------------------
    int iBornPWt1  = 0, iBornNWt1  = 0;
    int iBornPWtN1 = 0, iBornNWtN1 = 0;
    int iLoopPWt1  = 0, iLoopNWt1  = 0;
    int iLoopPWtN1 = 0, iLoopNWtN1 = 0;
//--------------------------------------------------------------------------------------------------
    mkdir("Results", S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
    output_born.open("Results/SANCeeLOnobr.lhe");
    output_loop.open("Results/SANCeeNLOnobr.lhe");
//--------------------------------------------------------------------------------------------------
    print_born_init();
    print_loop_init();
//--------------------------------------------------------------------------------------------------
    TFile *fborn = new TFile("Results/SANCeeLOnobr.root","recreate");
    TTree *tborn = new TTree("events_born","born events");
    Double_t weight_born[1], px_born[3], py_born[3], pz_born[3], E_born[3];
    tborn->Branch("weight_born", weight_born, "weight_born[1]/D");
    tborn->Branch("px_born", px_born, "px_born[3]/D");
    tborn->Branch("py_born", py_born, "py_born[3]/D");
    tborn->Branch("pz_born", pz_born, "pz_born[3]/D");
    tborn->Branch("E_born", E_born, "E_born[3]/D");
//--------------------------------------------------------------------------------------------------
    for (loop = 0; loop < NevBorn; loop++) {
	FoamBorn->MakeEvent();   // generate Born MC event
	FoamBorn->GetMCvect(MCvectBorn);
	MCwt = FoamBorn->GetMCwt();
	if (MCwt == 1.0) {
	    if (PrintLHE == 1) {
		print_born_lhe();
	    }
	    xwgtup = hepeup_.xwgtup;
	    weight_born[0] = hepeup_.xwgtup;
	    px_born[0] = hepeup_.pup[2][0];
	    px_born[1] = hepeup_.pup[3][0];
	    px_born[2] = hepeup_.pup[4][0];
	    py_born[0] = hepeup_.pup[2][1];
	    py_born[1] = hepeup_.pup[3][1];
	    py_born[2] = hepeup_.pup[4][1];
	    pz_born[0] = hepeup_.pup[2][2];
	    pz_born[1] = hepeup_.pup[3][2];
	    pz_born[2] = hepeup_.pup[4][2];
	    E_born[0]  = hepeup_.pup[2][3];
	    E_born[1]  = hepeup_.pup[3][3];
	    E_born[2]  = hepeup_.pup[4][3];
	    if (PrintROOT == 1) {
		tborn->Fill();
	    }
	    if (xwgtup > 0.0) iBornPWt1++;
	    if (xwgtup < 0.0) iBornNWt1++;
	}
	if (MCwt > 1.0) {
	    xwgtup = hepeup_.xwgtup;
	    if (xwgtup > 0.0) iBornPWtN1++;
	    if (xwgtup < 0.0) { iBornNWtN1++; MCwt = -MCwt; }
	    cout << "Born " << "MCwt   = " << SP12 << MCwt << endl;
	}
    }
    fborn->Write();
//--------------------------------------------------------------------------------------------------
    TFile *floop = new TFile("Results/SANCeeNLOnobr.root","recreate");
    TTree *tloop = new TTree("events_loop","loop events");
    Double_t weight_loop[1], px_loop[3], py_loop[3], pz_loop[3], E_loop[3];
    tloop->Branch("weight_loop", weight_loop, "weight_loop[1]/D");
    tloop->Branch("px_loop", px_loop, "px_loop[3]/D");
    tloop->Branch("py_loop", py_loop, "py_loop[3]/D");
    tloop->Branch("pz_loop", pz_loop, "pz_loop[3]/D");
    tloop->Branch("E_loop", E_loop, "E_loop[3]/D");
//--------------------------------------------------------------------------------------------------
    for (loop = 0; loop < NevLoop; loop++) {
	FoamLoop->MakeEvent();   // generate Loop MC event
	FoamLoop->GetMCvect(MCvectLoop);
	MCwt = FoamLoop->GetMCwt();
	if (MCwt == 1.0) {
	    if (PrintLHE == 1) {
		print_loop_lhe();
	    }
	    xwgtup = hepeup_.xwgtup;
	    weight_loop[0] = hepeup_.xwgtup;
	    px_loop[0] = hepeup_.pup[2][0];
	    px_loop[1] = hepeup_.pup[3][0];
	    px_loop[2] = hepeup_.pup[4][0];
	    py_loop[0] = hepeup_.pup[2][1];
	    py_loop[1] = hepeup_.pup[3][1];
	    py_loop[2] = hepeup_.pup[4][1];
	    pz_loop[0] = hepeup_.pup[2][2];
	    pz_loop[1] = hepeup_.pup[3][2];
	    pz_loop[2] = hepeup_.pup[4][2];
	    E_loop[0]  = hepeup_.pup[2][3];
	    E_loop[1]  = hepeup_.pup[3][3];
	    E_loop[2]  = hepeup_.pup[4][3];
	    if (PrintROOT == 1) {
		tloop->Fill();
	    }
	    if (xwgtup > 0.0) iLoopPWt1++;
	    if (xwgtup < 0.0) iLoopNWt1++;
	}
	if (MCwt > 1.0) {
	    xwgtup = hepeup_.xwgtup;
	    if (xwgtup > 0.0) iLoopPWtN1++;
	    if (xwgtup < 0.0) { iLoopNWtN1++; MCwt = -MCwt; }
	    cout << "Loop " << "MCwt   = " << SP12 << MCwt << endl;
	}
    }
    floop->Write();
//--------------------------------------------------------------------------------------------------
    output_born << "</LesHouchesEvents>" << endl;
    output_loop << "</LesHouchesEvents>" << endl;
    output_born.close();
    output_loop.close();
//--------------------------------------------------------------------------------------------------
    double BornF = 1.0;
    double LoopF = 1.0;
    if ((iBornPWt1+iBornNWt1) > 0) BornF = (double) (iBornPWt1-iBornNWt1)/(iBornPWt1+iBornNWt1);
    if ((iLoopPWt1+iLoopNWt1) > 0) LoopF = (double) (iLoopPWt1-iLoopNWt1)/(iLoopPWt1+iLoopNWt1);
//--------------------------------------------------------------------------------------------------
    cout << "====== Events generated, entering Finalize" << endl;
    Double_t IntNorm, Errel;
    FoamBorn->Finalize(IntNorm, Errel);
    FoamLoop->Finalize(IntNorm, Errel);
//--------------------------------------------------------------------------------------------------
    FoamBorn->GetIntegMC(MCresult, MCerror);  // get MC integral
    SigmaBorn = BornF*MCresult;
    double ErrBorn = MCerror;
    double SigmaBornAbs = MCresult;
    cout << "================================================================" << endl;
    cout << "        MCresult (Born) = " << SigmaBorn << " +- " << ErrBorn << " RelErr = " << ErrBorn/SigmaBorn << endl;
    delete [] MCvectBorn;
//--------------------------------------------------------------------------------------------------
    FoamLoop->GetIntegMC(MCresult, MCerror);  // get MC integral
    SigmaLoop = LoopF*MCresult;
    double ErrLoop = MCerror;
    double SigmaLoopAbs = MCresult;
    cout << "        MCresult (Loop) = " << SigmaLoop << " +- " << ErrLoop << " RelErr = " << ErrLoop/SigmaLoop << endl;
    delete [] MCvectLoop;
//--------------------------------------------------------------------------------------------------
    double SigmaDel = SigmaLoop-SigmaBorn;
    double ErrDel = sqrt(ErrLoop*ErrLoop+ErrBorn*ErrBorn);
    double Delta = SigmaDel/SigmaBorn;
    double ErrDelta = Delta*sqrt(ErrDel/SigmaDel*ErrDel/SigmaDel+ErrBorn/SigmaBorn*ErrBorn/SigmaBorn);
    cout << "        MCresult (Delta)  = " << Delta << " +- " << ErrDelta << " RelErr = " << ErrDelta/Delta << endl;
    cout << "================================================================" << endl;
    ofstream sigma_born;
    ofstream sigma_loop;
    sigma_born.open("Results/SANCeeLOnobr.dat");
    sigma_loop.open("Results/SANCeeNLOnobr.dat");
    sigma_born << setprecision(10) << SigmaBornAbs << endl;
    sigma_loop << setprecision(10) << SigmaLoopAbs << endl;
    sigma_born.close();
    sigma_loop.close();
//--------------------------------------------------------------------------------------------------
    cout << "Total number of Born events: " << iBornPWt1 + iBornPWtN1 << " pos. " << iBornNWt1 + iBornNWtN1 << " neg." << endl;
    cout << "Total number of Loop events: " << iLoopPWt1 + iLoopPWtN1 << " pos. " << iLoopNWt1 + iLoopNWtN1 << " neg." << endl;
    cout << "================================================================" << endl;
    cout << "Number of Born events with Wt > 1: " << iBornPWtN1 << " pos. " << iBornNWtN1 << " neg." << endl;
    cout << "Number of Loop events with Wt > 1: " << iLoopPWtN1 << " pos. " << iLoopNWtN1 << " neg." << endl;
    cout << "================================================================" << endl;
    cout << "Relative number of Born events with Wt > 1: " << (double) (iBornPWtN1+iBornNWtN1)/(iBornPWt1+iBornNWt1+iBornPWtN1+iBornNWtN1) << endl;
    cout << "Relative number of Loop events with Wt > 1: " << (double) (iLoopPWtN1+iLoopNWtN1)/(iLoopPWt1+iLoopNWt1+iLoopPWtN1+iLoopNWtN1) << endl;
    cout << "***** End of Program *****" << endl;
    return 0;
//-- end of main program ---------------------------------------------------------------------------
}
