# Find the Pythia8 includes and library.
#
# This module defines
# PYTHIA8_INCLUDE_DIR, where to locate Pythia.h file
# PYTHIA8_LIBRARIES, the libraries to link against to use Pythia8
# PYTHIA8_FOUND.  If false, you cannot build anything that requires Pythia8.

message(STATUS "Looking for Pythia8")

set(PYTHIA8_FOUND 0)

#message(STATUS "Found PYTHIA8_INCLUDE_DIR at ${PYTHIA8_INCLUDE_DIR}")

#set(CMAKE_FIND_LIBRARY_SUFFIXES ".a;.so")
find_library(PYTHIA8_LIBRARY
    NAMES
        pythia8 libpythia8.la libpythia8.a libpythia8.so
    PATHS
        $ENV{PYTHIA8_ROOT}
        $ENV{PYTHIA8}
        ${PYTHIA8_DIR}
        /opt/pythia8
        /usr/local
        /usr
        ~/local
        ~/local/pythia8
    PATH_SUFFIXES archive lib lib64
    DOC "Specify the Pythia8 library here."
    NO_DEFAULT_PATH
)
message(STATUS "Found PYTHIA8_LIBRARY at ${PYTHIA8_LIBRARY}")

find_library(PYTHIA8_LHAPDF6_LIBRARY
    NAMES
        libpythia8lhapdf6.so pythia8lhapdf6 libpythia8lhapdf6.a libpythia8lhapdf6.la
    PATHS
        $ENV{PYTHIA8_ROOT}
        $ENV{PYTHIA8}
        ${PYTHIA8_DIR}
        /opt/pythia8
        /usr/local
        /usr
        ~/local
        ~/local/pythia8
    PATH_SUFFIXES lib lib64
    DOC "Specify the Pythia8lhapdf6 library here."
)

message(STATUS "Found PYTHIA8_LHAPDF6_LIBRARY at ${PYTHIA8_LHAPDF6_LIBRARY}")

STRING(REGEX REPLACE "(^.*)/lib[pP]ythia8(.a|.so|.la)?$" "\\1" PYTHIA8_LIBRARY_DIR ${PYTHIA8_LIBRARY})
STRING(REGEX REPLACE "(^.*)/lib(/archive)?$" "\\1" PYTHIA8_DIR ${PYTHIA8_LIBRARY_DIR})
#message(STATUS "Found PYTHIA8_LIBRARY_DIR at ${PYTHIA8_LIBRARY_DIR}")

if(LHAPDF_LIBRARY)
    set(PYTHIA8_LIBRARIES ${PYTHIA8_LIBRARY} ${PYTHIA8_LHAPDF6_LIBRARY} -ldl)
else(LHAPDF_LIBRARY)
    set(PYTHIA8_LIBRARIES ${PYTHIA8_LIBRARY} ${PYTHIA8_LIBRARY_DIR}/liblhapdfdummy.a)
endif(LHAPDF_LIBRARY)

#STRING(REGEX REPLACE "(^.*)/include$" "\\1" PYTHIA8_DIR ${PYTHIA8_INCLUDE_DIR})
set(PYTHIA8_INCLUDE_DIR ${PYTHIA8_DIR}/include)
#message(STATUS "Found PYTHIA8_DIR at ${PYTHIA8_DIR}")

find_path(CHECK_PYTHIA8_INCLUDE_DIR Pythia.h
    PATHS
        ${PYTHIA8_INCLUDE_DIR}
    PATH_SUFFIXES Pythia8
    DOC "Check Pythia.h."
)

find_path(PYTHIA8_XMLDOC_DIR Index.xml
    PATHS
        ${PYTHIA8_DIR}
        /usr/share/doc/packages/pythia/xmldoc
    PATH_SUFFIXES xmldoc share/Pythia8/xmldoc
    DOC "Search xmldoc dir."
)

if(CHECK_PYTHIA8_INCLUDE_DIR AND PYTHIA8_LIBRARY)
  set(PYTHIA8_FOUND 1 )
  message(STATUS "Found Pythia8 library at ${PYTHIA8_LIBRARY}")
  message(STATUS "With xmldoc at ${PYTHIA8_XMLDOC_DIR}")
endif()

MARK_AS_ADVANCED( PYTHIA8_FOUND PYTHIA8_LIBRARY PYTHIA8_INCLUDE_DIR PYTHIA8_XMLDOC_DIR)
